/*
 * Title:        sag_connectivity_threading.hpp
 * Description:  A cross-platform threading implementation for old compilers
 * $Copyright (c) 2015-2016, 2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: sag_connectivity_threading.h 366779 2020-02-05 10:42:04Z matj $
 */

/** @file
 *
 * Provides macros for threading and concurrency primitives.
 *
 * Note that you should only use these if your compiler is too old to support
 * C++11 threading primitives. On modern compilers you should use 
 * std::thread, std::mutex, std::unique_lock, std::condition_variable etc.
 *
 * These macros are defined for both windows and linux to allow creation of a
 * cross-platform plugin. If you don't need your plugin to work on multiple
 * platforms then you can use platform-specific threading mechanisms if easier.
 *
 * This file defines three types:
 *
 * - SAG_MUTEX_T - mutexe
 * - SAG_SEMA_T - semaphors
 * - SAG_THREAD_T - threa.
 *
 * There are functions to init and destroy mutex and semaphore variables:
 *
 * - mutex - SAG_MUTEX_INIT(m) and SAG_MUTEX_DESTROY(m)
 * - semaphore - SAG_SEMA_INIT(s) and SAG_SEMA_DESTROY(s)
 *
 * The mutex/semaphore-specific functions:
 *
 * - mutex - SAG_MUTEX_LOCK(m) and SAG_MUTEX_UNLOCK(m)
 * (but see also com::softwareag::connectivity::SAG_LOCK_GUARD which provides a more exception-safe/RAII way to lock and unlock)
 * - semaphore - SAG_SEMA_WAIT(s) and SAG_SEMA_POST(s)
 *
 * Threads are created with:
 *
 * SAG_THREAD_T id;
 * SAG_THREAD_CREATE(id, func, args)
 *
 * func must match this type:
 *
 * SAG_THREAD_RET_T func(void* args)
 * {
 *		...
 *		SAG_THREAD_RETURN;
 * }
 *
 * You can join a running thread with SAG_THREAD_JOIN(id);
 *
 */

#ifndef _SAG_CONNECTIVITY_THREADS_HPP_
#define _SAG_CONNECTIVITY_THREADS_HPP_

#if defined(_WIN32) || defined(__WIN32) || defined(WIN32)

// Windows implementation using CriticalSections and _beginthreadex

#include <stdarg.h>
#include <windef.h>
#include <winbase.h>
#include <process.h>
#include <io.h>

/** Lock a mutex. In most situations com::softwareag::connectivity::SAG_LOCK_GUARD should be used instead of calling this directly. */
#define SAG_MUTEX_LOCK(m) EnterCriticalSection(&m)
/** Unlock a mutex. In most situations com::softwareag::connectivity::SAG_LOCK_GUARD should be used instead of calling this directly. */
#define SAG_MUTEX_UNLOCK(m) LeaveCriticalSection(&m)
/** Create a mutex.  */
#define SAG_MUTEX_INIT(m) InitializeCriticalSection(&m)
/** Destroy a mutex. */
#define SAG_MUTEX_DESTROY(m) DeleteCriticalSection(&m)
/** A helper that provides cross-platform mutex functionality for use on old compilers.
 *
 * Note that you should only use this if your compiler is too old to support
 * C++11 threading primitives. On modern compilers you should use std::mutex.
 */
#define SAG_MUTEX_T CRITICAL_SECTION

/** Wait/acquire a semaphore. */
#define SAG_SEMA_WAIT(s) WaitForSingleObject(s, INFINITE)
/** Post/signal/release a semaphore. */
#define SAG_SEMA_POST(s) ReleaseSemaphore(s, 1, NULL)
/** Create a semaphore. */
#define SAG_SEMA_INIT(s) s = CreateSemaphore(NULL, 0, 0x7FFFFFFF, NULL)
/** Destroy a semaphore. */
#define SAG_SEMA_DESTROY(s) CloseHandle(s)
/** A helper that provides cross-platform semaphore functionality for use on old compilers.
*
* Note that you should only use this if your compiler is too old to support
* C++11 threading primitives. On modern compilers you should use 
* std::mutex and std::condition_variable.
*/
#define SAG_SEMA_T HANDLE

/** Create a thread which will call func with params. id will be written with the thread ID
*
* Note that you should only use this if your compiler is too old to support
* C++11 threading primitives. On modern compilers you should use std::thread.
*/
#define SAG_THREAD_CREATE(id, func, params) (!(id = (HANDLE)_beginthreadex(NULL, 0, func, params, 0, NULL)))
/** Wait the given thread to terminate */
#define SAG_THREAD_JOIN(id) WaitForSingleObject(id, INFINITE)
/** The statement to return from a thread function */
#define SAG_THREAD_RETURN return 0;
/** The type of a thread */
#define SAG_THREAD_T HANDLE 
/** The return type of a thread function */
#define SAG_THREAD_RET_T unsigned int __stdcall 

#else // defined(_WIN32) || defined(__WIN32) || defined(WIN32)

// Unix implementation using pthreads

#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>
#include <signal.h>


/** Lock a mutex. In most situations com::softwareag::connectivity::SAG_LOCK_GUARD should be used instead of calling this directly. */
#define SAG_MUTEX_LOCK(m) pthread_mutex_lock(&m)
/** Unlock a mutex. In most situations com::softwareag::connectivity::SAG_LOCK_GUARD should be used instead of calling this directly. */
#define SAG_MUTEX_UNLOCK(m) pthread_mutex_unlock(&m)
/** Create a mutex.  */
#define SAG_MUTEX_INIT(m) pthread_mutex_init(&m, NULL)
/** Destroy a mutex. */
#define SAG_MUTEX_DESTROY(m) pthread_mutex_destroy(&m)
/** A helper that provides cross-platform mutex functionality for use on old compilers.
*
* Note that you should only use this if your compiler is too old to support
* C++11 threading primitives. On modern compilers you should use std::mutex.
*/

#define SAG_MUTEX_T pthread_mutex_t

/** Wait/acquire a semaphore. */
#define SAG_SEMA_WAIT(s) {while (sem_wait(&s));}
/** Post/signal/release a semaphore. */
#define SAG_SEMA_POST(s) sem_post(&s)
/** Create a semaphore. */
#define SAG_SEMA_INIT(s) sem_init(&s, 0, 0)
/** Destroy a semaphore. */
#define SAG_SEMA_DESTROY(s) sem_destroy(&s)
/** A helper that provides cross-platform semaphore functionality for use on old compilers.
*
* Note that you should only use this if your compiler is too old to support
* C++11 threading primitives. On modern compilers you should use 
* std::mutex and std::condition_variable.
*/
#define SAG_SEMA_T sem_t

/** Create a thread which will call func with params. id will be written with the thread ID
 *
 * Note that you should only use this if your compiler is too old to support
 * C++11 threading primitives. On modern compilers you should use std::thread.
 */
#define SAG_THREAD_CREATE(id, func, params) pthread_create(& id, NULL, func, params)
/** Wait the given thread to terminate */
#define SAG_THREAD_JOIN(id) pthread_join(id, NULL)
/** The statement to return from a thread function */
#define SAG_THREAD_RETURN return NULL;
/** The type of a thread */
#define SAG_THREAD_T pthread_t
/** The return type of a thread function */
#define SAG_THREAD_RET_T void*


#endif // (else) defined(_WIN32) || defined(__WIN32) || defined(WIN32)

#ifdef __cplusplus

#include <stdexcept>

namespace com {
namespace softwareag {
namespace _DATAT_INTERNAL_CPP_NAMESPACE {

/**
A simple RAII wrapper for SAG_MUTEX_T that locks the specified mutex in its constructor,
and guarantees that it will be unlocked when this object goes out of scope
unless it has been unlocked already by that point.

This is safer than relying on explicitly calling unlock() which can lead to hangs if 
an exception is thrown before the mutex is unlocked.

Note that you should only use this if your compiler is too old to support
C++11 threading primitives. On modern compilers you should use 
std::mutex with either std::lock_guard or std::unique_lock.
*/
class SAG_LOCK_GUARD
{
public:
	/** Create a SAG_LOG_GUARD for a particular mutex. Takes the mutex. */
	SAG_LOCK_GUARD(SAG_MUTEX_T &mx)
		:mx(mx), locked(false)
	{
		lock();
	}

	/** Destroy this lock guard. Releases the mutex if still held. */
	~SAG_LOCK_GUARD()
	{ 
		if (locked)
			unlock(); 
	}

	/** Locks the mutex (throws if the mutex is already locked by this lock guard). */
	void lock() 
	{
		if (locked) throw std::runtime_error("Cannot lock a SAG_MUTEX_T that is already locked");
		locked = true;
		SAG_MUTEX_LOCK(mx);
	}
	/** Unlocks the mutex (throws if the mutex is already unlocked by this lock guard). */
	void unlock() 
	{
		if (!locked) throw std::runtime_error("Cannot unlock a SAG_MUTEX_T that is already unlocked");
		locked = false;
		SAG_MUTEX_UNLOCK(mx);
	}
private:
	/** not copy-constructible */
	SAG_LOCK_GUARD(const SAG_LOCK_GUARD &other);
	/** not move-constructible */
	SAG_LOCK_GUARD(SAG_LOCK_GUARD &&other);
	/** nor assignable */
	SAG_LOCK_GUARD& operator=(const SAG_LOCK_GUARD &other);
	/** nor move-assignable */
	SAG_LOCK_GUARD& operator=(SAG_LOCK_GUARD &&other);

	/** reference to the mutex */
	SAG_MUTEX_T &mx;
	/** True if we hold the lock */
	bool locked;
};

}
namespace connectivity { using namespace _DATAT_INTERNAL_CPP_NAMESPACE; }
}} // com.softwareag.connectivity

#endif // __cplusplus

#endif // _SAG_CONNECTIVITY_THREADS_HPP_
